import uuid
from queue import Queue
from queue import Empty
import threading

lock = threading.Lock()


class UniqueKeyQueue:

    def __init__(self, ack=False):
        self._queue = Queue()
        self._queue_set = set()
        self._dict = {}
        self._ack = ack

    def put(self, key, data):
        try:
            lock.acquire()
            if key not in self._queue_set:
                _qid = str(uuid.uuid1())
                self._queue.put((data, _qid, key))
                self._queue_set.add(key)
                self._dict[_qid] = data
                return _qid
            return None
        finally:
            lock.release()

    def put_batch(self, data_list):
        qid_list = [self.put(key, data) for key, data in data_list]
        return qid_list

    def get(self, block=False, timeout=None):
        try:
            lock.acquire()
            data, qid, key = self._queue.get(block=block, timeout=timeout)
            self._queue_set.remove(key)
            if not self._ack:
                del self._dict[qid]
            return data, qid
        except Empty:
            return None, None
        finally:
            lock.release()

    def get_batch(self, block=False, timeout=None, num=1000):
        data_list = []
        try:
            while len(data_list) < num:
                data_list.append(self._queue.get(block=block, timeout=timeout))
        except Empty:
            pass
        return data_list

    def ack(self, qid):
        if not self._ack:
            return True
        if qid is None:
            return False
        del self._dict[qid]
        return True

    def queue_size(self):
        return self._queue.qsize()

    def set_size(self):
        return len(self._queue_set)

    def dict_size(self):
        return len(self._dict)


import time
from queue import Empty


def getn_from_queue(queue, n, sleep_seconds=0.0001):
    """
    通用的在queue里面批量取n个值再返回，方便后续的批量操作
    """
    results = []
    try:
        while len(results) < n:
            results.append(queue.get(block=False))
    except Empty:
        time.sleep(sleep_seconds)
    return results



from multiprocessing.managers import BaseManager
from .common.protocol import DEFAULT_SIMPLE_AGENT_NAME
from .common.protocol import DEFAULT_UNIQUE_KEY_AGENT_NAME
from .clients.simple_agent_client import SimpleAgentClient
from .clients.unique_key_agent_client import UniqueKeyAgentClient
from loguru import logger

class ClientFactory:

	def __init__(self, host, port, authkey):
		logger.debug(f'client factory init host={host}, port={port}')
		BaseManager.register(DEFAULT_SIMPLE_AGENT_NAME)
		BaseManager.register(DEFAULT_UNIQUE_KEY_AGENT_NAME)
		m = BaseManager(address=(host, port), authkey=authkey)
		m.connect()
		self._unique_key_manager = getattr(m, DEFAULT_UNIQUE_KEY_AGENT_NAME)()
		self._simple_manager = getattr(m, DEFAULT_SIMPLE_AGENT_NAME)()

	def create_simple_client(self, topic_name):
		simple_client = SimpleAgentClient(self._simple_manager, topic_name)
		return simple_client

	def create_unique_key_client(self, topic_name):
		unique_key_client = UniqueKeyAgentClient(self._unique_key_manager, topic_name)
		return unique_key_client
